' ------------------------------------------------------------------------------------------------
' Milkyway Type
' ------------------------------------------------------------------------------------------------
Type TMilkyway

	' Variables
	Field Stars:Int = 1000										' Number of fixed stars
	Field MaxAngle:Float = 45									' Maximum angle of Star creation
	Field Scale:Float = 512										' Scale of Scene
	Field Height:Float = Scale * 0.75							' Height of Panorama
	Field Segments:Int = 8										' Panorama Mesh Segments
	Field StarScale:Float = 1.5									' Star size multiplicator
	Field FixedStars:Float = 0.25								' Amount of larger stars (0...1)
	
	' Colors
	Field RGB1:Int[] = [0, 0, 0, 0]								' Panorama upper vertex colors
	Field RGB2:Int[] = [255, 255, 255, 128]						' Panorama center vertex colors
	Field RGB3:Int[] = [0, 0, 0, 0]								' Panorama lower vertex colors
	
	' Nebulae
	Field NebulaAlpha:Float = 0.125
	Field NebulaRows:Int = 4
	Field NebulaCols:Int = 4
	Field NebulaMinSize:Int = 128
	Field NebulaMedSize:Int = 256
	Field NebulaMaxSize:Int = 512
	Field NebulaMaxAngle:Float = 20
	Field NebulaAmount:Int = 50
	Field NebulaDensity:Float = 0.25							' Density of larger Nebulas

	' Entities
	Field Pivot:TPivot
	Field Starbox:TMesh
	Field Panorama:TMesh
	Field StarSphere:TMesh
	Field Nebula:TMesh
	
	' Textures
	Field StarTexture:TTexture
	Field StarsTexture:TTexture
	Field PanoramaTexture:TTexture
	Field NebulaTexture:TTexture
	
	Field st:TQuad

	
	
	' --------------------------------------------------------------------------------------------
	' Milkyway Type
	' --------------------------------------------------------------------------------------------
	Method Init()
					
		' Main Pivot
		Pivot = CreatePivot()
			
		' Create Meshes
		InitStarBox(Pivot)										' Create Background Stars
		InitPanorama(Pivot)										' Create Panorama
		InitNebula(Pivot)										' Create Nebulae
		InitStarSphere(Pivot)									' Create Front Stars
				
		' Reorder
		EntityOrder StarSphere, 1
		EntityOrder Nebula, 2
		EntityOrder Panorama, 3
		EntityOrder Starbox, 4
		
		' Texture Scaling
		ScaleTexture NebulaTexture, NebulaRows, NebulaCols
			
		' Entity FX
		EntityFX Starbox, 1
		EntityFX Panorama, 1 + 2 + 32
		EntityFX StarSphere, 1 + 2 + 32
		EntityFX Nebula, 1 + 2 + 32
			
		' Entity Blending
		EntityBlend Starbox, 3
		EntityBlend Panorama, 3
		EntityBlend StarSphere, 3
		EntityBlend Nebula, 3

		' Entity Texturing
		EntityTexture Panorama, PanoramaTexture
		EntityTexture Starbox, StarsTexture
		EntityTexture StarSphere, StarTexture
		EntityTexture Nebula, NebulaTexture
			
	End Method



	' --------------------------------------------------------------------------------------------
	' Update Panorama Vertex Colors
	' --------------------------------------------------------------------------------------------
	Method UpdatePanoramaColor()
	
		Local surf:TSurface = GetSurface(Panorama, 1)
				
		For Local v:Int = 0 To CountVertices(surf) - 1 Step 6

			VertexColor surf, v + 0, RGB1[0], RGB1[1], RGB1[2], RGB1[3] / 255.0
			VertexColor surf, v + 1, RGB1[0], RGB1[1], RGB1[2], RGB1[3] / 255.0
			VertexColor surf, v + 2, RGB2[0], RGB2[1], RGB2[2], RGB2[3] / 255.0
			VertexColor surf, v + 3, RGB2[0], RGB2[1], RGB2[2], RGB2[3] / 255.0
			VertexColor surf, v + 4, RGB3[0], RGB3[1], RGB3[2], RGB3[3] / 255.0
			VertexColor surf, v + 5, RGB3[0], RGB3[1], RGB3[2], RGB3[3] / 255.0
		
		Next
			
	End Method

	
			
	' --------------------------------------------------------------------------------------------
	' Background Starbox
	' --------------------------------------------------------------------------------------------
	Method InitStarBox(parent:TEntity = Null)
	
		Starbox = CreateCube(parent)
		FlipMesh Starbox
		ScaleEntity Starbox, Scale, Scale, Scale
			
	End Method



	' --------------------------------------------------------------------------------------------
	' Milkyway Panorama
	' --------------------------------------------------------------------------------------------
	Method InitPanorama(parent:TEntity = Null)
		
		Local a1:Float, a2:Float, inc:Float, x1:Float, z1:Float, x2:Float, z2:Float
		Local uu1:Float, uu2:Float, uv1:Float, uv2:Float, uv3:Float
		Local v0:Int, v1:Int, v2:Int, v3:Int, v4:Int, v5:Int
		Local surf:TSurface
			
		Panorama = CreateMesh(parent)
		surf = CreateSurface(Panorama)
			
		inc = 360.0 / Segments
			
		While a1 < 360
				
			a2 = (a1 + inc) Mod 360
					
			x1 = Scale * Cos(a1)
			z1 = Scale * Sin(a1)
			x2 = Scale * Cos(a2)
			z2 = Scale * Sin(a2)

			uu1 = 1.0 - (a1 / 360.0)
			uu2 = 1.0 - (a1 + inc) / 360.0

			uv1 = 1.00
			uv2 = 0.50
			uv3 = 0.00
			
			v0 = AddVertex(surf, x1, -Height, z1, uu1, uv1)
			v1 = AddVertex(surf, x2, -Height, z2, uu2, uv1)
			v2 = AddVertex(surf, x1, 0, z1, uu1, uv2)
			v3 = AddVertex(surf, x2, 0, z2, uu2, uv2)
			v4 = AddVertex(surf, x1, Height, z1, uu1, uv3)
			v5 = AddVertex(surf, x2, Height, z2, uu2, uv3)
			
			VertexColor surf, v0, RGB1[0], RGB1[1], RGB1[2], RGB1[3] / 255.0
			VertexColor surf, v1, RGB1[0], RGB1[1], RGB1[2], RGB1[3] / 255.0
			VertexColor surf, v2, RGB2[0], RGB2[1], RGB2[2], RGB2[3] / 255.0
			VertexColor surf, v3, RGB2[0], RGB2[1], RGB2[2], RGB2[3] / 255.0
			VertexColor surf, v4, RGB3[0], RGB3[1], RGB3[2], RGB3[3] / 255.0
			VertexColor surf, v5, RGB3[0], RGB3[1], RGB3[2], RGB3[3] / 255.0
			
			AddTriangle surf, v0, v1, v3
			AddTriangle surf, v3, v2, v0
			AddTriangle surf, v2, v5, v4
			AddTriangle surf, v5, v2, v3
			
			a1:+inc
					
		Wend
				
	End Method

	
	
	' --------------------------------------------------------------------------------------------
	' Fixed Stars
	' --------------------------------------------------------------------------------------------
	Method InitStarSphere:TMesh(parent:TEntity = Null)
		
		Local i:Int
		Local q:TQuad
		Local r:Int, g:Int, b:Int, a:Float
		Local surf:TSurface
		Local col:Float
		Local size:Float
		
		StarSphere = CreateMesh(parent)
		surf = CreateSurface(StarSphere)
						
		For i = 1 To Stars
		
			st = New TQuad
			st.mesh = StarSphere
			st.surf = surf
			st.scalex = Rnd(1.5, 2.5) * StarScale
			
			If Rnd(1) > (1 - FixedStars) Then st.scalex = Rnd(2.5, 5.0) * StarScale
			
			st.scaley = st.scalex
			
			' color
			col = Rnd(1)
			
			' 25% orange stars
			If col > 0.75 And col <= 1.00 Then
			
				r = 255
				g = 224
				b = 192
				
			EndIf
			
			' 25% yellow stars
			If col > 0.50 And col <= 0.75 Then
			
				r = 255
				g = 255
				b = 192
				
			EndIf
			
			' 25% blue stars	
			If col > 0.25 And col <= 0.50 Then
			
				r = 192
				g = 224
				b = 255
				
			EndIf
			
			' 25% white stars
			If col > 0.00 And col <= 0.25 Then
			
				r = 255
				g = 255
				b = 255
				
			EndIf
				
			st.RGB1 = [r, g, b]
			st.RGB2 = [r, g, b]
			st.RGB3 = [r, g, b]
			st.RGB4 = [r, g, b]
			
			st.x = Rnd(-100, 100)
			st.y = Rnd(-MaxAngle, MaxAngle)
			st.z = Rnd(-100, 100)
					
			TFormNormal st.x, st.y, st.z, Null, Null
			
			Local d:Float = 1'Rnd(0.5, 1.5)
			
			st.x = TFormedX() * Scale * d
			st.y = TFormedY() * Scale * d
			st.z = TFormedZ() * Scale * d
				
			st.Add()
			st.AlignToTarget(st.x, st.y, st.z)
			
		Next
		
	End Method

	
		
	' --------------------------------------------------------------------------------------------
	' Detail Nebula
	' --------------------------------------------------------------------------------------------
	Method InitNebula(parent:TEntity = Null)
	
		Local i:Int
		Local q:TQuad
		Local r:Int, g:Int, b:Int, a:Float
		Local surf:TSurface
		
		Nebula = CreateMesh(parent)
		surf = CreateSurface(Nebula)
						
		For i = 1 To NebulaAmount
		
			Local q:TQuad = New TQuad
			q.mesh = Nebula
			q.surf = surf
			q.scalex = Rnd(NebulaMinSize, NebulaMedSize)
			q.scaley = q.scalex
			
			r = Rand(128, 255)
			g = Rand(128, 255)
			b = Rand(128, 255)
				
			q.RGB1 = [r, g, b]
			q.RGB2 = [r, g, b]
			q.RGB3 = [r, g, b]
			q.RGB4 = [r, g, b]
						
			q.Alpha1 = Rnd(1) * NebulaAlpha
			q.Alpha2 = Rnd(1) * NebulaAlpha
			q.Alpha3 = Rnd(1) * NebulaAlpha
			q.Alpha4 = Rnd(1) * NebulaAlpha
			
			Local num:Int = Rand(0, (NebulaRows * NebulaCols) - 1)
			Local row:Int = Int(num) / NebulaRows
			Local col:Int = Int(num) - (row * NebulaCols)
			
			q.x = Rnd(-100, 100)
			q.y = Rnd(-NebulaMaxAngle, NebulaMaxAngle)
			q.z = Rnd(-100, 100)
			
			If Rnd(1) > (1 - NebulaDensity) Then
				
				q.y = Rnd(-100, 100)
				q.scalex = Rnd(NebulaMedSize, NebulaMaxSize)
				q.scaley = q.scalex
		
				r = Rand(0, 255)
				g = Rand(0, 255)
				b = Rand(0, 255)
					
				q.RGB1 = [r, g, b]
				q.RGB2 = [r, g, b]
				q.RGB3 = [r, g, b]
				q.RGB4 = [r, g, b]
					
			End If
			
			TFormNormal q.x, q.y, q.z, Null, Null
						
			q.x = TFormedX() * Scale
			q.y = TFormedY() * Scale
			q.z = TFormedZ() * Scale
				
			q.Add(row, col)
			q.AlignToTarget(q.x, q.y, q.z, Rnd(360))
			
		Next	
	
	End Method
	

	
	' --------------------------------------------------------------------------------------------
	' Hold position
	' --------------------------------------------------------------------------------------------
	Method KeepPosition(parent:TEntity)

		PositionEntity Pivot, EntityX(parent, 1), EntityY(parent, 1), EntityZ(parent, 1)

	End Method



End Type