'Const QuatToEulerAccuracy:Double = 1.0 / 2 ^ 512

Type TQuat
	
	Field w:Double, x:Double, y:Double, z:Double
	
End Type

Function EulerToQuat(out:TQuat, pitch:Double, yaw:Double, roll:Double)
	
	' convert a Rotation to a Quat
	' NB roll is inverted due to change in handedness of coordinate systems
	Local cr:Double = Cos(-roll / 2)
	Local cp:Double = Cos(pitch / 2)
	Local cy:Double = Cos(yaw / 2)
	
	Local sr:Double = Sin(-roll / 2)
	Local sp:Double = Sin(pitch / 2)
	Local sy:Double = Sin(yaw / 2)
	
	' These variables are only here to cut down on the number of multiplications
	Local cpcy:Double = cp * cy
	Local spsy:Double = sp * sy
	Local spcy:Double = sp * cy
	Local cpsy:Double = cp * sy
	
	' Generate the output quat
	out.w = cr * cpcy + sr * spsy
	out.x = sr * cpcy - cr * spsy
	out.y = cr * spcy + sr * cpsy
	out.z = cr * cpsy - sr * spcy
	
End Function

Function QuatToEntity(src:TQuat, Entity:TEntity, RotGlobal:Int = 0)
	
	' convert a Quat to a Rotation
	
	Local sint:Double, cost:Double, sinv:Double, cosv:Double, sinf:Double, cosf:Double
	Local cost_temp:Double
	
	sint = (2 * src.w * src.y) - (2 * src.x * src.z)
	cost_temp = 1.0 - (sint * sint)
	
	If Abs(cost_temp) > QuatToEulerAccuracy
		
		cost = Sqr(cost_temp)
		
	Else
		
		cost = 0
		
	EndIf
	
	If Abs(cost) > QuatToEulerAccuracy
		
		sinv = ((2 * src.y * src.z) + (2 * src.w * src.x)) * 1.0 / cost
		cosv = (1 - (2 * src.x * src.x) - (2 * src.y * src.y)) * 1.0 / cost
		sinf = ((2 * src.x * src.y) + (2 * src.w * src.z)) * 1.0 / cost
		cosf = (1 - (2 * src.y * src.y) - (2 * src.z * src.z)) * 1.0 / cost
		
	Else
		
		sinv = (2 * src.w * src.x) - (2 * src.y * src.z)
		cosv = 1 - (2 * src.x * src.x) - (2 * src.z * src.z)
		sinf = 0
		cosf = 1
		
	EndIf
	
	' Generate the output rotation, inverted due to change in handedness of coordinate system
	
	Local a1:Float = ATan2(sint, cost)
	Local a2:Float = ATan2(sinf, cosf)
	Local a3:Float = -ATan2(sinv, cosv)
	
	if a1<>Null then RotateEntity Entity, a1, a2, a3, RotGlobal
	
End Function

Function QuatToEntity2(src:TQuat, Entity:TEntity, RotGlobal:Int = 0)
	
	' convert a Quat to a Rotation
	
	Local sint:Double, cost:Double, sinv:Double, cosv:Double, sinf:Double, cosf:Double
	Local cost_temp:Double
	
	sint = (2 * src.w * src.y) - (2 * src.x * src.z)
	cost_temp = 1.0 - (sint * sint)
	
	If Abs(cost_temp) > QuatToEulerAccuracy
		
		cost = Sqr(cost_temp)
		
	Else
		
		cost = 0
		
	EndIf
	
	If Abs(cost) > QuatToEulerAccuracy
		
		sinv = ((2 * src.y * src.z) + (2 * src.w * src.x)) * 1.0 / cost
		cosv = (1 - (2 * src.x * src.x) - (2 * src.y * src.y)) * 1.0 / cost
		sinf = ((2 * src.x * src.y) + (2 * src.w * src.z)) * 1.0 / cost
		cosf = (1 - (2 * src.y * src.y) - (2 * src.z * src.z)) * 1.0 / cost
		
	Else
		
		sinv = (2 * src.w * src.x) - (2 * src.y * src.z)
		cosv = 1 - (2 * src.x * src.x) - (2 * src.z * src.z)
		sinf = 0
		cosf = 1
		
	EndIf
	
	' Generate the output rotation, inverted due to change in handedness of coordinate system
	
	Local a1:Float = ATan2(sint, cost)
	Local a2:Float = ATan2(sinf, cosf)
	Local a3:Float = -ATan2(sinv, cosv)
	
	if a1<>Null then RotateEntity Entity, a1, a2, a3, RotGlobal
	
End Function

Function MultiplyQuat(result:TQuat, q1:TQuat, q2:TQuat)
	
	' result will be the same rotation as doing q1 then q2 (order matters!)
	
	Local a:Double, b:Double, c:Double, d:Double, e:Double, f:Double, g:Double, h:Double
	
	a = (q1.w + q1.x) * (q2.w + q2.x)
	b = (q1.z - q1.y) * (q2.y - q2.z)
	c = (q1.w - q1.x) * (q2.y + q2.z)
	d = (q1.y + q1.z) * (q2.w - q2.x)
	e = (q1.x + q1.z) * (q2.x + q2.y)
	f = (q1.x - q1.z) * (q2.x - q2.y)
	g = (q1.w + q1.y) * (q2.w - q2.z)
	h = (q1.w - q1.y) * (q2.w + q2.z)
	
	result.w = b + (-e - f + g + h) / 2.0
	result.x = a - (e + f + g + h) / 2.0
	result.y = c + (e - f + g - h) / 2.0
	result.z = d + (e - f - g + h) / 2.0
	
End Function